<p>Hello {{ name }},</p>

<p>
  Thank you for joining ASReview. You are almost ready! Please click on the
  following link to confirm your account:
</p>

<p><a href="{{ url }}">{{ url }}</a></p>

<p>
  After confirmation you will be redirected to the signin page. Use your account
  credentials and start using ASReview.
</p>

<p>Regards,</p>

<p>The ASReview Team</p>
